/***                                                                      ***\

    logresolve 3.0


    Robert R. Collins / rcollins@x86.org / http://www.x86.org, May  1, 1998


    Usage: logresolve [arguments] < access_log > new_log        (v1.0 compliant)
    - or - logresolve [arguments] [< access_log] [> new_log]    (v2.0 compliant)

    Default Behavior:   Identical to logresolve version 1.0.
                        Input is taken from STDIN.
                        Output is directed to STDOUT.
                        Statistics are directed to STDERR.

    Arguments (if specified):  specifying any other arguments causes
                               statistics to be printed to STDERR (or ERRFILE).
        -i INFILE  :    Indicates an input file to be read.  When this argument
                        is given, input is read from the input file, not STDIN.
                        This argument overrides the default behavior.

        -o OUTFILE :    Indicates an output file to be written.  When this argument
                        is given, output is directed to the output file, not STDOUT.
                        This argument overrides the default behavior.

        -co [CACHE_OUT] Indicates that statistics will be reported after logresolve
                        finishes processing the input file.  When this argument
                        is given, all statistics are directed to STDERR unless
                        the sub-argument ERRFILE is present.  When ERRFILE is
                        present, statistics are directed to the file specified
                        by ERRFILE.  This argument overrides the default
                        behavior.
        -ci [CACHE_IN]  Indicates that a cache file will be used to preload the
                        DNS cache.

    Notes:

    To generate meaningful statistics from an HTTPD log file, it's good
    to have the domain name of each machine that accessed your site, but
    doing this on the fly can slow HTTPD down.

    Compiling NCSA HTTPD with the -DMINIMAL_DNS flag turns IP#->hostname
    resolution off. Before running your stats program, just run your log
    file through this program (logresolve) and all of your IP numbers will
    be resolved into hostnames (where possible).

    logresolve takes an HTTPD access log (in the COMMON log file format,
    or any other format that has the IP number/domain name as the first
    field for that matter), and outputs the same file with all of the
    domain names looked up. Where no domain name can be found, the IP
    number is left in.

    To minimize impact on your nameserver, logresolve has its very own
    internal direct-mapped cache. This means that each IP number will only
    be looked up the first time it is found in the log file. As noted
    above, giving any command-line arguments to logresolve (anything at
    all!) will give you some statistics on the log file printed to STDERR.

\***                                                                      ***/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "log.h"
#include "logdata.h"

int
main(argc, argv)
    int             argc;
    char           *argv[];
{
    char            infile[MAXLINE]     = {""};
    char            outfile[MAXLINE]    = {""};
    char            errfile[MAXLINE]    = {""};
    char            cachefile[MAXLINE]  = {""};
    char            line[MAXLINE]       = {""};
    char            msg[MAXLINE]        = {""};
    int             i, j, lookups, response;
    int             shmid0, shmid1;
    int             blocks = 1;
    FILE           *in, *out, *err, *cache;
    pid_t           pid;

    in  = stdin;
    out = stdout;
    err = stderr;

    for(i = argc, argc = 1; argc < i; argc ++)
    {
        switch(argv[argc][0])
        {
            case '-' :
                switch(argv[argc][1])
                {
                    case 'i' :          /* Capture INPUT file           */
                        strcpy(infile, argv[++argc]);
                        Status.UseInFile    = 1;
                        Status.PrintStats   = 1;
                        continue;
    
                    case 'd' :          /* Debug Flag                   */
                        Status.Debug = 1;
                        continue;
    
                    case 'c' :          /* Capture DNS Cache file        */
                        switch(argv[argc][2])
                        {
                        case 'i' :
                            strcpy(cachefile, argv[++argc]);
                            Status.UseCacheFile = 1;
                            Status.PrintStats   = 1;
                            break;

                        case 'o' :
                            strcpy(errfile, argv[++argc]);
                            Status.UseErrFile   = 1;
                            Status.PrintStats   = 1;
                            break;

                        default :
                            printf("Invalid command-line option: %s.\n", argv[argc]);
                            exit(-1);
                        }
                        continue;

                    case 'o' :          /* Capture OUTPUT file           */
                        strcpy(outfile, argv[++argc]);
                        Status.UseOutFile = 1;
                        Status.PrintStats   = 1;
                        continue;

                    case '-' :                          /* Print help message.                  */
                    case 'h' :
                    case '?' :
                        printf("\n");
                        printf("logresolve -- Resolve IP address in a log file to domain names.\n\n");
                        printf("Usage:  logresolve -i infile -o outfile [-ci cachefile_in] [-co cachefile_out]\n");
                        printf("   -i infile        : Specify input file .\n");
                        printf("   -o outfile       : Specify output file.\n");
                        printf("  [-ci cachefile]   : Specify input DNS cache file.  The input file format\n");
                        printf("                      is the same as the output cache file format\n");
                        printf("                      (-co option).\n");
                        printf("  [-co cachefile]   : Specify output DNS cache file.  The output file format\n");
                        printf("                      is the same as the input cache file format\n");
                        printf("                      (-ci option).  The output cache file also contains\n");
                        printf("                      statistics about the last run of logresolve.\n");
                        printf("\n");
                        fflush(stdout);
                        return (0);
                }
                    
            default :          /* Capture INPUT file           */
                printf("Invalid command-line option: %s.\n", argv[argc]);
                exit(-1);
        }
    }

    /* Create a socket pair.                                            */
    if(socketpair(AF_UNIX, SOCK_DGRAM, 0, sp))
    {
        fprintf(stderr, "Unable to create socket pair.\n");
        fflush(stderr);
        exit(1);
    }


    /* If an input cachefile has been specified, but not an output      *
     * cachefile, then the output cache file should assume the same     *
     * name as the input cachefile.                                     */
    if(Status.UseCacheFile && !Status.UseErrFile)
    {
        strcpy(errfile, cachefile);
        Status.UseErrFile = 1;
    }

    /* Open the input and output files.  We can't open the cache files  *
     * here, because we might want to use the same filename for the     *
     * input and output DNS cache files.                                */
    if(Status.UseInFile)
    {
        if((in = fopen(infile,"r")) == NULL) 
        {
            fprintf(stderr,"Input file not found:  %s\n",infile);
            exit(1);
        }
    }

    if(Status.UseOutFile)
    {
        if((out = fopen(outfile,"w")) == NULL)
        {
            fprintf(stderr,"Can't open output file:  %s\n",outfile);
            exit(1);
        }
    }

    /* Signals are installed to allow interprocess communications.  The *
     * following signales are used for this purpose:                    *
     * SIGQUIT - Signals the child process to terminate.                */
    signal(SIGQUIT, sig_handler);

    /* Shared memory is needed to share common data between processes.  *
     * The following data structures are shared between all processes:  *
     * children - A structure array containing control information for  *
     *            each child process.                                   *
     * dnscache - The DNS cache structure must be shared among all      *
     *            child processes.                                      */
    shmid0 = shmget(IPC_PRIVATE, sizeof(struct ipc_s) * MAXFILLSIZE, IPC_CREAT | 0x1b0);
    if(shmid0 == -1)
    {
        fprintf(stderr, "Unable to create shared memory block for children.\n");
        exit(1);
    }
    children = (struct ipc_s *) shmat(shmid0, (void *)0, 0);
    shmid1 = shmget(IPC_PRIVATE, sizeof(struct dnsrec) * BUCKETS, IPC_CREAT | 0x1b0);
    if(shmid1 == -1)
    {
        shmctl(shmid0, IPC_RMID, &shmdat);
        fprintf(stderr, "Unable to create shared memory block for DNS cache.\n");
        exit(1);
    }
    dnscache = (struct dnsrec *) shmat(shmid1, (void *)0, 0);

    /* Initialize the DNS cache.  All "evict" records in the DNS cache  *
     *will be initialized at this point.                                */
    for(i = 0; i < BUCKETS; i ++)
    {
        dnscache[i].evicts = -1;
    }

    /* Preload the DNS cache, if needed.                                */
    if(Status.UseCacheFile)
    {
        if((cache = fopen(cachefile,"r")) == NULL)
        {
            fprintf(stderr,"Can't open input DNS cache file:  %s\n",cachefile);
        }
        while(feof(cache) == 0)
        {
            if((fgets(line, sizeof line, cache)) == NULL)
                break;
            if(strstr(line, "BEGIN_DNS_CACHE") != NULL)
            {
                fgets(line, sizeof line, cache);
                break;
            }
        }
        while(feof(cache) == 0)
        {
            if((fgets(line, sizeof line, cache)) == NULL)
                break;
            if(strstr(line, "END_DNS_CACHE") != NULL)
            {
                break;
            }
            ParseCacheLine(line);
        }
        fclose(cache);
    }

    /* Create the child processes.  MAXCHILDREN child processes are     *
     * created for dispatching DNS lookups.  After reading an entry     *
     * from the log file, the parent process will attempt to resolve    *
     * name from our internal DNS cache.  If the name does not exist in *
     * the DNS cache,then the request will be dispatched to the child   *
     * process.                                                         */
    for(i = 0; i < MAXCHILDREN; i ++)
    {
        if((pid = fork()) == -1)
        {
            fprintf(stderr, "Unable to fork(%i).\n", i);
            kill(-getpid(), SIGQUIT);
            exit(1);
        }
        if(!pid)
        {
            logchild(i);
            children[i].msgin = MSG_CHILDTERMINATED;
            sprintf(msg, "Terminated child(%i) process:  %i.\n", i, getpid());
            Debug(msg);
            exit(0);
        }
        children[i].pid = pid;
        kill(children[i].pid, SIGCONT);
    }

    /* Process the input log file.  The input log file will be handled  *
     * in a psuedo nested loop.  MAXFILLSIZE lines are always read,     *
     * except when we reach the end-of-file.  During normal processing, *
     * the outer loop counter will always equal MAXFILLSIZE.  When EOF  *
     * is reached, our  outer loop counter indicates how many lines     *
     * remain in the input log file (modulo MAXFILLSIZE).  This simple  *
     * optimization makes end-of-file processing very easy, and without *
     * creating a special case to handle it.                            */
    while(feof(in) == 0)
    {
        /* getblock() retrieves a block of log file entries and returns *
         * the number of entries read.  Normally, the return will       *
         * always equal MAXFILLSIZE.  On the last block, this function  *
         * will return the residual amount of log file entires left in  *
         * the file.  All subsequent function calls use this return     *
         * value.  Therefore handling the end-of-file condition where   *
         * the last block of log entries read equals (modulo            *
         * MAXFILLSIZE) is quite easy.                                  */
        sprintf(msg, "%i. FEOF = %i; Calling getblock().\n", blocks, feof(in));
        Debug(msg);
        i = getblock(in);
        sprintf(msg, "%i. FEOF = %i; Returned with %i-lines from getblock().\n", blocks, feof(in), i);
        Debug(msg);

        /* From the block of log file entries just read, let's extract  *
         * all of the IP addresses.  All pre-resolved entries will have *
         * their IP address equal 0.0.0.0.  This is an illegal IP       *
         * address, therefore it makes a convenient semaphore for       *
         * subsequent processing.                                       */
        ExtractIPNUM(i);
        sprintf(msg, "%i. Returned from ExtractIPNUM().\n", blocks);
        Debug(msg);

        /* Sort entries by IP number.  All pre-resolved entries have    *
         * an IP number equal 0.0.0.0.  Therefore, all of these entries *
         * will remain at the beginning of the sort.                    */
        SortEntriesByIPNUM(i);
        sprintf(msg, "%i. Returned from SortEntriesByIPNUM().\n", blocks);
        Debug(msg);

        /* Look up the cached entries.  If an entry is cached, then the *
         * resolved domain name is copied to the first entry in the     *
         * "children" array.  All subsequent log entries with the same  *
         * IP number will assign a pointer to this original data.       */
        FindCachedEntries(i);
        sprintf(msg, "%i. Returned from FindCachedEntries().\n", blocks);
        Debug(msg);
        lookups = 0;
        for(j = 0; j < i; j ++)
        {
            switch(children[children[j].sortindex].msgin)
            {
            case LU_LOOKUP :
                lookups ++;
                break;

            case LU_EVICTION:
                lookups ++;
                break;
            }
        }

        sprintf(msg, "%i. Dispatching %i lookup requests.\n", blocks, lookups);
        Debug(msg);

        lookups = 0;
        for(j = 0; j < i; j ++)
        {
            switch(children[children[j].sortindex].msgin)
            {
            case LU_LOOKUP :
                lookups ++;
                send(sp[0], &children[j].sortindex, sizeof(children[j].sortindex), 0);
                break;

            case LU_EVICTION:
                lookups ++;
                send(sp[0], &children[j].sortindex, sizeof(children[j].sortindex), 0);
                sprintf(msg, "%i. Waiting for %i lookup responses.\n", blocks, lookups);
                Debug(msg);
                while(lookups)
                {
                    recv(sp[0], &response, sizeof(response), 0);
                    (response ? statistics.resolved ++ : statistics.noname ++);
                    lookups --;
                }
                break;

            case LU_IDLE:
                statistics.psuedo_hits ++;
                break;
            }
        }
        sprintf(msg, "%i. Waiting for %i lookup responses.\n", blocks, lookups);
        Debug(msg);
        while(lookups)
        {
            recv(sp[0], &response, sizeof(response), 0);
            (response ? statistics.resolved ++ : statistics.noname ++);
            lookups --;
        }
        sprintf(msg, "%i. Received all lookup responses.\n", blocks);
        Debug(msg);

        /* At last, we send our output to the correct output file.      *
         * If we didn't wait for all of the processes to register       *
         * their completion, then we would most certainly have written  *
         * our results out of order with respect to the input file.     */
        for(j = 0; j < i; j ++)
        {
            if(children[j].dnsname != NULL)
            {
                fprintf(out, "%s ", children[j].dnsname);
            }
            fprintf(out, "%s\n", children[j].lineout);
        }
        sprintf(msg, "%i. Done writing all entries to log file.\n", blocks);
        Debug(msg);
        blocks ++;

    }
    Debug("Done processing log file.\n");


    /* Now that we're done processing the file, let's kill all of the   *
     * child processes.                                                 */
    kill(-getpid(), SIGQUIT);
    Debug("Killed child processes.\n");

    /* Wait for all of the children to indicate their termination.      */
    for(i = 0; i < MAXCHILDREN; i ++)
    {
        waitpid(children[i].pid, NULL, WUNTRACED);
    }
    Debug("All children dead.\n");

    /* Terminate our shared memory.                                     */
    if(shmctl(shmid0, IPC_RMID, &shmdat))
    {
        fprintf(stderr, "Unable to terminate shared memory for children.\n");
    }
    if(shmctl(shmid1, IPC_RMID, &shmdat))
    {
        fprintf(stderr, "Unable to terminate shared memory for DNS cache.\n");
    }
    Debug("All shared memory terminated.\n");

    if(Status.PrintStats)
    {
        PrintStats(stderr);
    }
    Debug("Print stats done.\n");


    /* Close all the files, and split.                                  */
    if(Status.UseErrFile)
    {
        if((err = fopen(errfile,"w")) == NULL)
        {
            fprintf(stderr,"Can't open output DNS cache file:  %s\n",errfile);
            err = stderr;
        }
        stats(err);
        fclose(err);
    }
    Debug("Print stats to cache file done.\n");

    if(Status.UseInFile)
        fclose(in);
    Debug("Input file closed.\n");

    if(Status.UseOutFile)
        fclose(out);
    Debug("Output file closed.\n");

    return (0);
}
